@@ -15,11 +15,11 @@ PGObject::Util::BulkLoad - Bulk load records into PostgreSQL
 
 =head1 VERSION
 
-Version 0.03
+Version 0.04
 
 =cut
 
-our $VERSION = '0.03';
+our $VERSION = '0.04';
 
 
 =head1 SYNOPSIS
@@ -135,6 +135,10 @@ sql COPY statement
 
 Update/Insert CTE pulling temp table
 
+=item stats
+
+Get stats on pending upsert, grouped by an arbitrary column.
+
 =back
 
 =item table
@@ -157,6 +161,10 @@ Column names for update
 
 Names of columns in primary key.
 
+=item group_stats_by
+
+Names of columns to group stats by
+
 =back
 
 =cut
@@ -167,6 +175,36 @@ sub _sanitize_ident {
     qq("$string");
 }
 
+sub _statement_stats {
+    my ($args) = @_;
+    croak 'Key columns must array ref' unless (ref $args->{key_cols}) =~ /ARRAY/;
+    croak 'Must supply key columns' unless @{$args->{key_cols}};
+    croak 'Must supply table name' unless $args->{table};
+    croak 'Must supply temp table' unless $args->{tempname};
+
+    my @groupcols;
+    @groupcols = $args->{group_stats_by} 
+                 ? @{$args->{group_stats_by}} 
+                 : @{$args->{key_cols}};
+    my $table = _sanitize_ident($args->{table});
+    my $temp = _sanitize_ident($args->{tempname});
+    "SELECT " . join(', ', map {"$temp." . _sanitize_ident($_)} @groupcols) . ",
+            SUM(CASE WHEN ROW(" . join(', ', map {"$table." . _sanitize_ident($_)
+                                      } @{$args->{key_cols}}) . ") IS NULL
+                     THEN 1
+                     ELSE 0
+             END) AS pgobject_bulkload_inserts,
+            SUM(CASE WHEN ROW(" . join(', ', map {"$table." . _sanitize_ident($_)
+                                      } @{$args->{key_cols}}) . ") IS NULL
+                     THEN 0
+                     ELSE 1
+            END) AS pgobject_bulkload_updates
+       FROM $temp
+  LEFT JOIN $table USING (" . join(', ', map {_sanitize_ident($_)
+                                             } @{$args->{key_cols}}) . ")
+   GROUP BY " . join(', ', map {"$temp." . _sanitize_ident($_)} @groupcols);
+}
+
 sub _statement_temp {
     my ($args) = @_;
 
@@ -243,13 +281,22 @@ Columns to update (by name)
 
 Key columns (by name)
 
+=item group_stats_by
+
+This is an array of column names for optional stats retrieval and grouping.
+If it is set then we will grab the stats and return them.  Note this has a 
+performance penalty because it means an extra scan of the temp table and an
+extra join against the parent table.  See get_stats for the return value
+information if this is set.
+
 =back
 
 =cut
 
 sub _build_args {
     my ($init_args, $obj) = @_;
-    my @arglist = qw(table insert_cols update_cols key_cols dbh);
+    my @arglist = qw(table insert_cols update_cols key_cols dbh 
+                     tempname group_stats_by);
     return { 
        map {  my $val;
               for my $v ($init_args->{$_}, try { $obj->$_ } ){
@@ -274,14 +321,25 @@ sub upsert {
     # a permanent table there, they are inviting disaster.  At any rate this is
     # safe but a plain drop without schema qualification risks losing user data.
 
+    my $return_value;
+
     $dbh->do("DROP TABLE IF EXISTS pg_temp.pgobject_bulkloader");
     $dbh->do(statement( %$args, (type => 'temp', 
                               tempname => 'pgobject_bulkloader')
     ));
     copy({(%$args, (table => 'pgobject_bulkloader'))}, @_);
+
+    if ($args->{group_stats_by}){
+        $return_value = get_stats(
+                {(%$args, (tempname => 'pgobject_bulkloader'))}
+        );
+    }
+
     $dbh->do(statement( %$args, (type => 'upsert', 
                               tempname => 'pgobject_bulkloader')));
-    $dbh->do("DROP TABLE pg_temp.pgobject_bulkloader");
+    my $dropstatus = $dbh->do("DROP TABLE pg_temp.pgobject_bulkloader");
+    return $return_value if $args->{group_stats_by};
+    return $dropstatus;
 }
 
 =head2 copy
@@ -329,6 +387,63 @@ sub copy {
     $dbh->pg_putcopyend();
 }
 
+=head2 get_stats
+
+Takes the same arguments as upsert plus group_stats_by
+
+Returns an array of hashrefs representing the number of inserts and updates
+that an upsert will perform.  It must be performed before the upsert statement
+actually runs.  Typically this is run via the upsert command (which 
+automatically runs this if group_stats_by is set in the argumements hash).
+
+There is a performance penalty here since an unindexed left join is required 
+between the temp and the normal table.
+
+This function requires tempname, table, and group_stats_by to be set in the
+argument hashref.  The return value is a list of hashrefs with the following 
+keys:
+
+=over
+
+=item stats
+
+Hashref with keys inserts and updates including numbers of rows.
+
+=item keys
+
+Hashref for key columns and their values, by name
+
+=back
+
+=cut
+
+sub get_stats {
+    my ($args) = shift;
+    $args = shift if $args eq __PACKAGE__;
+    try {
+       no warnings;
+       no strict;
+       $args->can('foo');
+       unshift @_, $args; # args is an object
+    };
+    $args = _build_args($args, $_[0]);
+    my $dbh = $args->{dbh};
+
+    my $returnval = [
+          map { 
+            my @row = @$_;
+            { stats => {
+                  updates => pop @row,
+                  inserts => pop @row,
+              },
+              keys => {
+                 map { $_ => shift @row } @{$args->{group_stats_by}}
+              },
+            } 
+          } @{ $dbh->selectall_arrayref(statement(%$args, (type => 'stats'))) }
+    ];
+}
+
 =head1 AUTHOR
 
 Chris Travers, C<< <chris.travers at gmail.com> >>
